import numpy as np
import pdb

'''
NOTE: Need to recompute the spectra between each iteration (EUGH).

Steps:
    - For a given parameter, choose 4 even variations withing 1 sigma contour (i.e. this will probably be an input)
    - Calculate Fisher as dClobs/dparm
        - Calculate derivative with 5-point stencil method http://en.wikipedia.org/wiki/Five-point_stencil
    - Find the new 1 sigma contour according to this Fisher
    - Iterate until contours converge to some tolerance -- 0.5% as default

Inputs:
    - function to calculate vector of Clobs
Outpus:
    - Converged Fisher matrix

Write to parallelise
will be passed a pool with map function can pool.map(params) will give list of results
make list of params, run all at once, will parallelise nicely
be aware that some of the workers in pool might fail, return e.g. None
can have all paramers normalised to [0,1]

compute_vector(p) runs an entire pipeline on a set of parameters (i.e. needs to be done five times here) and returns vector, covmat where vector is Cl_*** and covmat is the necessary **INVERSE** covmat

can do pool.map(compute_vector, [p1,p2 etc]) where p1 is an array of the parameter values

len(start_vector) will give you the number of parameters we are varying, which is taken from the cosmosis values.ini file

'''

class FisherParameterError(Exception):
    def __init__(self, parameter_index):
        message = "Fisher Matrix likelihood function returned None for parameter: {}".format(parameter_index)
        super(Exception,self).__init__(self, message)
        self.parameter_index = parameter_index


class Fisher(object):
    def __init__(self, compute_vector, start_vector, step_size, tolerance, maxiter, pool=None):
        
        self.compute_vector = compute_vector
        self.maxiter = maxiter
        self.step_size = step_size
        self.start_params = start_vector
        self.current_params = start_vector
        self.nparams = start_vector.shape[0]
        self.iterations = 0
        self.pool = pool

    def converged(self):
        crit = (abs(self.new_onesigma - self.old_onesigma).max() < self.threshold)
        return crit

    def converge_fisher_matrix(self):
        
        self.new_Fmatrix = compute_fisher_matrix(self.compute_vector,
            self.start_vector, self.start_covmat)
        
        self.old_onesigma = compute_one_sigma(new_Fmatrix)

        while True:
            self.iterations+=1
            self.old_onesigma = self.new_onesigma
            self.current_params = self.choose_new_params(self.new_Fmatrix)

            self.new_Fmatrix = self.compute_fisher_matrix()

            self.new_onesigma = compute_one_sigma(self.new_Fmatrix)

            if self.converged():
                print('Fisher has converged!')
                return new_Fmatrix

            if self.iterations > self.maxiter:
                print("Run out of iterations.")
                print("Done %d, max allowed %d" % (self.iterations, self.maxiter))
                return None

    def compute_derivatives(self):
        derivatives = []
        points = []

        #To improve parallelization we first gather all the data points
        #we use in all the dimensions
        for p in range(self.nparams):
            points +=  self.five_points_stencil_points(p)
        print("Calculating derivatives using {} total models".format(len(points)))
        if self.pool is None:
            results = list(map(self.compute_vector, points))
        else:
            results = self.pool.map(self.compute_vector, points)

        # Bit of a waste of time to compute the inv cov separately,
        # but it's a quick fix to a memory error if we compute the cov
        # for every single value
        _, inv_cov = self.compute_vector(points[0], cov=True)

        #Now get out the results that correspond to each dimension
        for p in range(self.nparams):
            results_p = results[4*p:4*(p+1)]
            derivative = self.five_point_stencil_deriv(results_p, p)
            derivatives.append(derivative)
        derivatives = np.array(derivatives)
        return derivatives, inv_cov


    def compute_fisher_matrix(self):
        derivatives, inv_cov = self.compute_derivatives()

        if not np.allclose(inv_cov, inv_cov.T):
            print("WARNING: The inverse covariance matrix produced by your pipeline")
            print("         is not symmetric. This probably indicates a mistake somewhere.")
            print("         If you are only using cosmosis-standard-library likelihoods please ")
            print("         open an issue about this on the cosmosis site.")
        fisher_matrix = np.einsum("il,lk,jk->ij", derivatives, inv_cov, derivatives)
        return fisher_matrix

    def five_points_stencil_points(self, param_index):
        delta = np.zeros(self.nparams)
        delta[param_index] = 1.0
        points = [self.current_params + x*delta for x in 
            [2*self.step_size, 
             1*self.step_size, 
            -1*self.step_size, 
            -2*self.step_size]
        ]
        return points        

    def five_point_stencil_deriv(self, obs, param_index):
        for r in obs:
            if r is None:
                raise FisherParameterError(param_index)
        deriv = (-obs[0] + 8*obs[1] - 8*obs[2] + obs[3])/(12*self.step_size)
        return deriv

    def compute_one_sigma(Fmatrix):
        sigma = np.sqrt(np.linalg.inv(Fmatrix))
        return sigma

class NumDiffToolsFisher(Fisher):
    def compute_derivatives(self):
        import numdifftools as nd
        def wrapper(param_vector):
            print("Running pipeline:", param_vector)
            return self.compute_vector(param_vector, cov=False)
        jacobian_calculator = nd.Jacobian(wrapper, step=self.step_size)
        derivatives = jacobian_calculator(self.current_params)
        _, inv_cov = self.compute_vector(self.current_params, cov=True)
        print(derivatives.shape, inv_cov.shape)
        return derivatives.T, inv_cov
    


def test():
    def theory_prediction(x, cov=False):
        #same number of data points as parameters here
        x = np.concatenate([x,x])
        theory = 2*x + 2
        inv_cov = np.diag(np.ones_like(x)**-1)
        if cov:
            return theory, inv_cov
        else:
            return theory

    best_fit_params = np.array([0.1, 1.0, 2.0, 4.0,])
    fisher_calculator = Fisher(theory_prediction, best_fit_params, 0.01, 0.0, 100)
    F = fisher_calculator.compute_fisher_matrix()
    print(F)
    return F

if __name__ == '__main__':
    test()


from cosmosis.runtime.config import Inifile
from cosmosis.main import run_cosmosis, parser
import tempfile
import os
import sys
import pytest
import numpy as np


def test_sampler_chain():


    with tempfile.TemporaryDirectory() as dirname:
        values_file = f"{dirname}/values.txt"
        maxlike_file = f"{dirname}/chain.maxlike.txt"
        fisher_file = f"{dirname}/chain.fisher.txt"
        emcee_file = f"{dirname}/chain.txt"
        with open(values_file, "w") as values:
            values.write(
                "[parameters]\n"
                "p1=-3.0  0.0  3.0\n"
                "p2=-3.0  0.0  3.0\n")

        params = {
            ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
            ('runtime', 'sampler'):  "maxlike fisher emcee",
            ("pipeline", "debug"): "T",
            ("pipeline", "modules"): "test1",
            ("pipeline", "extra_output"): "parameters/p3",
            ("pipeline", "values"): values_file,
            ("test1", "file"): "example_module.py",
            ("output", "filename"): emcee_file,
            ("emcee", "walkers"): "8",
            ("emcee", "samples"): "100",
            ("maxlike", "tolerance"): "0.05",
            ("fisher", "step_size"): "0.01"
        }


        ini = Inifile(None, override=params)

        status = run_cosmosis(ini)

        data = np.loadtxt(fisher_file)
        print(data.shape)

        data = np.loadtxt(maxlike_file)
        print(data.shape)

        data = np.loadtxt(emcee_file)
        print(data.shape)



    assert status == 0




from cosmosis import Inifile, Sampler, LikelihoodPipeline, InMemoryOutput, TextColumnOutput
from cosmosis.postprocessing import postprocessor_for_sampler
import tempfile
import os
import sys
import pytest
import numpy as np
from astropy.table import Table


minuit_compiled = os.path.exists(Sampler.get_sampler("minuit").libminuit_name)

def run(name, check_prior, check_extra=True, can_postprocess=True, do_truth=False, no_extra=False, pp_extra=True, pp_2d=True, **options):

    sampler_class = Sampler.registry[name]

    values = tempfile.NamedTemporaryFile('w')
    values.write(
        "[parameters]\n"
        "p1=-3.0  0.0  3.0\n"
        "p2=-3.0  0.0  3.0\n")
    values.flush()

    override = {
        ('runtime', 'root'): os.path.split(os.path.abspath(__file__))[0],
        ("pipeline", "debug"): "F",
        ("pipeline", "modules"): "test1",
        ("pipeline", "extra_output"): "parameters/p3",
        ("pipeline", "values"): values.name,
        ("test1", "file"): "example_module.py",
    }

    for k,v in options.items():
        override[(name,k)] = str(v)

    if no_extra:
        del override[("pipeline", "extra_output")]


    ini = Inifile(None, override=override)

    # Make the pipeline itself
    pipeline = LikelihoodPipeline(ini)

    output = InMemoryOutput()
    sampler = sampler_class(ini, pipeline, output)
    sampler.config()


    while not sampler.is_converged():
        sampler.execute()

    if check_prior:
        pr = np.array(output['prior'])
        # a few samples might be outside the bounds and have zero prior
        assert np.all((pr==-3.58351893845611)|(pr==-np.inf))
        # but not all of them
        assert not np.all(pr==-np.inf)

    if check_extra and not no_extra:
        p1 = output['parameters--p1']
        p2 = output['parameters--p2']
        p3 = output['PARAMETERS--P3']
        assert np.all((p1+p2==p3)|(np.isnan(p3)))

    if can_postprocess:
        pp_class = postprocessor_for_sampler(name)
        print(pp_class)
        with tempfile.TemporaryDirectory() as dirname:
            truth_file = values.name if do_truth else None
            pp = pp_class(output, "Chain", 0, outdir=dirname, prefix=name, truth=truth_file, fatal_errors=True)
            pp_files = pp.run()
            pp.finalize()
            pp.save()
            for p in pp_files:
                print(p)
            postprocess_files = ['parameters--p1', 'parameters--p2']
            if pp_2d:
                postprocess_files.append('2D_parameters--p2_parameters--p1')
            if check_extra and pp_extra and not no_extra:
                postprocess_files.append('parameters--p3')
                if pp_2d:
                    postprocess_files += ['2D_parameters--p3_parameters--p2', '2D_parameters--p3_parameters--p1']
            for p in postprocess_files:
                filename = f"{dirname}{os.path.sep}{name}_{p}.png"
                print("WANT ", filename)
                assert filename in pp_files
                assert os.path.exists(filename)
            for p in os.listdir(dirname):
                p = os.path.join(dirname, p)
                if p.endswith(".txt"):
                    Table.read(p, format='ascii.commented_header')
                    print(f"Read file {p} as a table")


    return output


def test_apriori():
    run('apriori', True, can_postprocess=False, nsample=100)

def test_dynesty():
    # dynesty does not support extra params
    run('dynesty', False, check_extra=False, nlive=50, sample='unif')

def test_emcee():
    run('emcee', True, walkers=8, samples=100)
    run('emcee', True, walkers=8, samples=100, a=3.0)

def test_truth():
    run('emcee', True, walkers=8, samples=100, do_truth=True)

def test_fisher():
    run('fisher', False, check_extra=False)

def test_grid():
    run('grid', True, pp_extra=False, nsample_dimension=10)

def test_gridmax():
    run('gridmax', True, can_postprocess=False, max_iterations=1000)

# def test_kombine():
#     run('kombine')

def test_maxlike():
    run('maxlike', True, can_postprocess=False)

def test_metropolis():
    run('metropolis', True, samples=20)
    run('metropolis', True, samples=20, covmat_sample_start=True)

@pytest.mark.skipif(not minuit_compiled,reason="requires Minuit2")
def test_minuit():
    run('minuit', True, can_postprocess=False)

def test_multinest():
    run('multinest', True, max_iterations=10000, live_points=50, feedback=False)

def test_pmaxlike():
    run('pmaxlike', True, can_postprocess=False)

def test_pmc():
    old_settings = np.seterr(invalid='ignore', divide='ignore')
    run('pmc', True, iterations=10)
    np.seterr(**old_settings)  

def test_zeus():
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50)
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50, verbose=True)
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50, tune=False)
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50, tolerance=0.1)
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50, patience=5000)
    run('zeus', True, maxiter=100_000, walkers=10, samples=100, nsteps=50, moves="differential:2.0  global")
    run('zeus', True, maxiter=50_000, walkers=10, samples=100, nsteps=50)

def test_polychord():
    with tempfile.TemporaryDirectory() as base_dir:
        run('polychord', True, live_points=20, feedback=0, base_dir=base_dir, polychord_outfile_root='pc')

def test_snake():
        run('snake', True, pp_extra=False)

def test_nautilus():
    run('nautilus', True)
    run('nautilus', True, n_live=500, enlarge_per_dim=1.05,
        split_threshold=95., n_networks=3, n_batch=50, verbose=True, f_live=0.02, n_shell=100)


def test_star():
        run('star', False, pp_extra=False, pp_2d=False)

def test_test():
    run('test', False, can_postprocess=False)

def test_list_sampler():
    # test that the burn and thin parameters work
    # make a mock output file with some samples
    with tempfile.TemporaryDirectory() as dirname:

        input_chain_filename = os.path.join(dirname, "input_chain.txt")
        input_chain = TextColumnOutput(input_chain_filename)
        input_chain.add_column('parameters--p1', float)
        input_chain.add_column('parameters--p2', float)
        nrow = 100
        data = np.random.normal(size=(nrow, 2)).clip(-2.99, 2.99)
        for i in range(nrow):
            input_chain.parameters(data[i, 0], data[i, 1])
        input_chain.close()

        burn = 0.2
        thin = 2

        data = data[20::thin]

        output = run("list", True, can_postprocess=False, filename=input_chain_filename, burn=burn, thin=thin)
        p1 = output['parameters--p1']
        p2 = output['parameters--p2']
        p3 = p1 + p2
        expected_like = -(p1**2 + p2**2)/2.
        assert p1.size == 40
        assert np.allclose(p1, data[:, 0])
        assert np.allclose(p2, data[:, 1])
        assert np.allclose(output['PARAMETERS--P3'], p3)
        assert np.allclose(output["post"] - output["prior"], expected_like)

@pytest.mark.skipif(sys.version_info < (3,7), reason="pocomc requires python3.6+")
def test_poco():
    try:
        import pocomc
    except ImportError:
        pytest.skip("pocomc not installed")
    run('pocomc', True, check_extra=False, n_effective=32, n_active=16,  n_total=32, n_evidence=32, )

if __name__ == '__main__':
    import sys
    locals()[sys.argv[1]]()


#coding: utf-8
from .. import ParallelSampler
from ...runtime import logs
import ctypes as ct
import os
import numpy as np
import sys
from ...runtime.utils import mkdir
import warnings

prior_type = ct.CFUNCTYPE(None, 
    ct.POINTER(ct.c_double),  #hypercube
    ct.POINTER(ct.c_double),  #physical
    ct.c_int,   #ndim
)

loglike_type = ct.CFUNCTYPE(ct.c_double, 
    ct.POINTER(ct.c_double),  #physical
    ct.c_int,   #ndim
    ct.POINTER(ct.c_double),  #derived
    ct.c_int,   #nderived
)


dumper_type = ct.CFUNCTYPE(None, #void
    ct.c_int,  #ndead
    ct.c_int,  #nlive
    ct.c_int,  #npars
    ct.POINTER(ct.c_double),   #live
    ct.POINTER(ct.c_double),   #dead
    ct.POINTER(ct.c_double),   #logweights
    ct.c_double,   #logZ
    ct.c_double,   #logZerr
)


polychord_args = [
    loglike_type,           #loglike,
    prior_type,             #prior,
    dumper_type,            #dumper,
    ct.c_int,               #nlive
    ct.c_int,               #nrepeats
    ct.c_int,               #nprior
    ct.c_bool,              #do_clustering
    ct.c_int,               #feedback
    ct.c_double,            #precision_criterion
    ct.c_double,            #logzero
    ct.c_int,               #max_ndead
    ct.c_double,            #boost_posterior
    ct.c_bool,              #posteriors
    ct.c_bool,              #equals
    ct.c_bool,              #cluster_posteriors
    ct.c_bool,              #write_resume 
    ct.c_bool,              #write_paramnames
    ct.c_bool,              #read_resume
    ct.c_bool,              #write_stats
    ct.c_bool,              #write_live
    ct.c_bool,              #write_dead
    ct.c_bool,              #write_prior
    ct.c_double,            #compression_factor
    ct.c_int,               #nDims
    ct.c_int,               #nDerived 
    ct.c_char_p,            #base_dir
    ct.c_char_p,            #file_root
    ct.c_int,               #nGrade
    ct.POINTER(ct.c_double),#grade_frac
    ct.POINTER(ct.c_int),   #grade_dims
    ct.c_int,               #n_nlives
    ct.POINTER(ct.c_double),#loglikes
    ct.POINTER(ct.c_int),   #nlives
    ct.c_int,               #seed
]


POLYCHORD_SECTION='polychord'


class PolychordSampler(ParallelSampler):
    parallel_output = False
    sampler_outputs = [("prior", float), ("like", float), ("post", float), ("weight", float)]
    supports_smp=False
    internal_resume = True
    understands_fast_subspaces = True

    def config(self):
        if self.pool:
            libname = "libchord_mpi.so"
        else:
            libname = "libchord.so"

        dirname = os.path.split(__file__)[0]
        libname = os.path.join(dirname, "polychord_src", libname)
            
        try:
            libchord = ct.cdll.LoadLibrary(libname)
        except Exception as error:
            sys.stderr.write("PolyChord could not be loaded.\n")
            sys.stderr.write("This may mean an MPI compiler was not found to compile it,\n")
            sys.stderr.write("or that some other error occurred.  More info below.\n")
            sys.stderr.write(str(error)+'\n')
            sys.exit(1)

        self._run = libchord.polychord_c_interface
        self._run.restype=None
        self._run.argtypes = polychord_args
        self.converged=False

        self.ndim = len(self.pipeline.varied_params)
        # We save the prior as well as the other derived params
        self.nderived = self.pipeline.number_extra + 1

        #Required options
        self.live_points    = self.read_ini("live_points", int, 100)

        #Output and feedback options
        self.feedback               = self.read_ini("feedback", int, 1)
        self.resume                 = self.read_ini("resume", bool, False)
        self.polychord_outfile_root = self.read_ini("polychord_outfile_root", str, "")
        self.base_dir               = self.read_ini("base_dir", str, ".")
        self.compression_factor     = self.read_ini("compression_factor", float, np.exp(-1))

        if len(self.polychord_outfile_root)>299:
            raise ValueError("Polychord parameter 'polychord_outfile_root'"
                " cannot be longer than 299s characters.")

        if '/' in self.polychord_outfile_root:
            warnings.warn("\n\nWARNING: / found in polychord_outfile_root\n"
                "In polychord it is better to set base_dir to the directory and \n"
                "have polychord_outfile_root not include the directory.\n"
                "Otherwise you may have to mkdir some directories yourself.\n"
                "This may cause problems below.\n"
                )

        #General run options
        self.max_iterations = self.read_ini("max_iterations", int, -1)
        self.num_repeats = self.read_ini("num_repeats", int, 0)
        self.nprior = self.read_ini("nprior", int, -1)
        self.random_seed = self.read_ini("random_seed", int, -1)
        self.tolerance   = self.read_ini("tolerance", float, 0.1)
        self.log_zero    = self.read_ini("log_zero", float, -1e6)
        self.boost_posteriors = self.read_ini("boost_posteriors", float, 0.0)
        self.weighted_posteriors = self.read_ini("weighted_posteriors", bool, True)
        self.equally_weighted_posteriors = self.read_ini("equally_weighted_posteriors", bool, True)       
        self.cluster_posteriors = self.read_ini("cluster_posteriors", bool, True)

        self.fast_fraction    = self.read_ini("fast_fraction", float, 0.5)

        if self.output:
            def dumper(ndead, nlive, npars, live, dead, logweights, log_z, log_z_err):
                logs.overview("Saving %d samples" % ndead)
                self.output_params(ndead, nlive, npars, live, dead, logweights, log_z, log_z_err)
            self.wrapped_output_logger = dumper_type(dumper)
        else:
            def dumper(ndead, nlive, npars, live, dead, logweights, log_z, log_z_err):
                pass
            self.wrapped_output_logger = dumper_type(dumper)

        def prior(cube, theta, ndim):
            cube_vector = np.array([cube[i] for i in range(ndim)])
            if self.pipeline.do_fast_slow:
                cube_vector = self.reorder_slow_fast(cube_vector)
            try:
                theta_vector = self.pipeline.denormalize_vector_from_prior(cube_vector) 
            except ValueError:
                # Polychord sometimes seems to propose outside the prior.
                # Just give terrible parameters when that happens.
                theta_vector = np.repeat(-np.inf, ndim)
            for i in range(ndim):
                theta[i] = theta_vector[i]

        self.wrapped_prior = prior_type(prior)

        def likelihood(theta, ndim, phi, nderived):
            theta_vector = np.array([theta[i] for i in range(ndim)])

            if np.any(~np.isfinite(theta_vector)):
                return -np.inf

            try:
                r = self.pipeline.run_results(theta_vector)

                # Extract the derived parameters
                for i in range(nderived-1):
                    phi[i] = r.extra[i]
                # And we also handle the prior here
                phi[nderived-1] = r.prior

            except KeyboardInterrupt:
                raise sys.exit(1)

            return r.like
        self.wrapped_likelihood = loglike_type(likelihood)

    def reorder_slow_fast(self, x):
        y = np.zeros_like(x)
        ns = self.pipeline.n_slow_params
        y[self.pipeline.slow_param_indices] = x[0:ns]
        y[self.pipeline.fast_param_indices] = x[ns:]
        return y



    def worker(self):
        self.sample()

    def execute(self):
        self.log_z = 0.0
        self.log_z_err = 0.0


        self.sample()

        self.output.final("log_z", self.log_z)
        self.output.final("log_z_error", self.log_z_err)

    def sample(self):

        if self.pipeline.do_fast_slow and (self.pipeline.n_fast_params > 0):
            logs.overview("Using two grades of parameter speed in polychord.")
            n_grade = 2
        elif self.pipeline.do_fast_slow:
            logs.warning("You asked for fast/slow, but there were no fast parameters, so "
                  "I have switched off Polychord's fast/slow mechanism to avoid a hang")
            n_grade = 1
        else:
            logs.overview("Using a single grade of parameter speeds in polychord.")
            n_grade = 1

        grade_dims = (ct.c_int*n_grade)()
        grade_frac = (ct.c_double*n_grade)()


        if n_grade > 1:
            if self.pipeline.n_slow_params == 0:
                raise ValueError("All your parameters have been classified as fast.  This will now work in polychord.  Change your fast/slow settings.")

            grade_dims[0] = self.pipeline.n_slow_params
            grade_dims[1] = self.pipeline.n_fast_params
            grade_frac[0] = 1 - self.fast_fraction
            grade_frac[1] = self.fast_fraction
            logs.overview("Telling Polychord to spend fraction {} if its time in the fast subspace (adjust with fast_fraction option)".format(self.fast_fraction))
        else:
            grade_dims[0] = self.pipeline.nvaried
            grade_frac[0] = 1.0


        n_nlives = 0
        loglikes = (ct.c_double*n_nlives)()
        nlives = (ct.c_int*n_nlives)()

        base_dir = self.base_dir.encode('ascii')
        polychord_outfile_root = self.polychord_outfile_root.encode('ascii')
        output_to_file = len(polychord_outfile_root) > 0

        if output_to_file:
            mkdir(self.base_dir)
            mkdir(os.path.join(self.base_dir, "clusters"))
            
        if self.num_repeats == 0:
            num_repeats = 3 * grade_dims[0]
            logs.overview("Polychord num_repeats = {}  (3 * n_slow_params [{}])".format(num_repeats, grade_dims[0]))
        else:
            num_repeats = self.num_repeats
            logs.overview("Polychord num_repeats = {}  (from parameter file)".format(num_repeats))

        self._run(
                self.wrapped_likelihood,      #loglike,
                self.wrapped_prior,           #prior,
                self.wrapped_output_logger,   #dumper,
                self.live_points,             #nlive
                num_repeats,                  #nrepeats
                self.nprior,                  #nprior
                True,                         #do_clustering
                self.feedback,                #feedback
                self.tolerance,               #precision_criterion
                self.log_zero,                #logzero
                self.max_iterations,          #max_ndead
                self.boost_posteriors,         #boost_posterior
                self.weighted_posteriors,     #posteriors
                self.equally_weighted_posteriors, #equals
                self.cluster_posteriors,      #cluster_posteriors
                True,                         #write_resume  - always
                False,                        #write_paramnames
                self.resume,                  #read_resume
                output_to_file,  #write_stats
                output_to_file,  #write_live
                output_to_file,  #write_dead
                output_to_file,  #write_prior
                self.compression_factor,      #compression_factor
                self.ndim,                    #nDims
                self.nderived,                #nDerived 
                base_dir,           #base_dir
                polychord_outfile_root,  #file_root
                n_grade,                      #nGrade
                grade_frac,                   #grade_frac
                grade_dims,                   #grade_dims
                n_nlives,                     #n_nlives
                loglikes,                     #loglikes
                nlives,                       #nlives
                self.random_seed,             #seed
                )

        self.converged = True

    def output_params(self, ndead, nlive, npars, live, dead, logweights, log_z, log_z_err):
        # Polychord repeats output, but with changed weights, so reset to the start
        # of the chain to overwrite them.
        self.output.reset_to_chain_start()
        self.log_z = log_z
        self.log_z_err = log_z_err
        data = np.array([dead[i] for i in range(npars*ndead)]).reshape((ndead, npars))
        logw = np.array([logweights[i] for i in range(ndead)])
        for row, w in zip(data,logw):
            params = row[:self.ndim]
            extra_vals = row[self.ndim:self.ndim+self.nderived-1]
            prior = row[self.ndim+self.nderived-1]
            birth_like = row[self.ndim+self.nderived]
            like = row[self.ndim+self.nderived+1]
            importance = np.exp(w)
            post = like + prior
            self.output.parameters(params, extra_vals, prior, like, post, importance)
        self.output.final("nsample", ndead)
        self.output.flush()

    def is_converged(self):
        return self.converged

# legacy alias
PolyChordSampler = PolychordSampler



import numpy as np

def setup(options):
    return {}

def execute(block, config):
    p3 = block['parameters', 'p3']
    p4 = block['parameters', 'p4']
    like = -(p3**2 + p4**2)/2.
    block['likelihoods', 'test3_like'] = like
    return 0



from cosmosis.output.text_output import TextColumnOutput
from cosmosis.output.fits_output import FitsOutput
from cosmosis.runtime.config import Inifile
import os


def read_input(filename, force_text=False, weighted=False):
    """
    Read cosmosis output data, either by:
     - specifying a cosmosis .txt output file
     - specifying a cosmosis .ini input file that includes the output file specification
     - specifying a directory containing cosmosis test sampler output
     - specifying a non-cosmosis output file containing samples or weighted samples
     - passing in an astropy table of the chain

    Params:
        filename: string, paths to any of the options described above
        force_text: bool, default=False - regardless of the file ending assume it is text columns
        weighted: bool, default=False - if a non-cosmosis file is passed, assume that its samples have weights
    Returns:
        sampler - string with the name of the sampler in
        ini - a dictionary of information containing the output and metadata, suitable for instantiating a postprocessor.
    """
    if "astropy" in str(type(filename)):
        from astropy.table import Table
        if not isinstance(filename, Table):
            raise ValueError("If you pass in an astropy table it must be an astropy table")
        metadata = filename.meta
        sampler = metadata.get("sampler")
        # in this case the table already contains all the information we need
        ini = filename

    elif filename.endswith("txt") or force_text:
        output_info = TextColumnOutput.load_from_options({"filename":filename})
        metadata=output_info[2][0]
        sampler = metadata.get("sampler")
        if sampler is None:
            print("This is not a cosmosis output file.")
            print("So I will assume it is a generic MCMC file")
            if weighted:
                sampler = "weighted_metropolis"
            else:
                sampler = "metropolis"
            ini = output_info
        else:
            ini = {"sampler":sampler, sampler:metadata, "data":output_info, "output":dict(format="text", filename=filename)}

    elif filename.endswith("fits"):
        output_info = FitsOutput.load_from_options({"filename":filename})
        metadata=output_info[2][0]
        sampler = metadata.get("sampler")
        if sampler is None:
            print("This is not a cosmosis output file.")
            print("So I will assume it is a generic MCMC file")
            if weighted:
                sampler = "weighted_metropolis"
            else:
                sampler = "metropolis"
            ini = output_info
        else:
            ini = {"sampler":sampler, sampler:metadata, "data":output_info, "output":dict(format="fits", filename=filename)}

    elif os.path.isdir(filename):
        ini = Inifile(None)
        ini.add_section("runtime")
        ini.add_section("test")
        sampler = "test"
        ini.set("runtime", "sampler", sampler)
        ini.set("test", "save_dir", filename)
    else:
        #Determine the sampler and get the class
        #designed to postprocess the output of that sampler
        ini = Inifile(filename)
        sampler = ini.get("runtime", "sampler")

    return sampler, ini



from .. import ParallelSampler, sample_ellipsoid, sample_ball
from ...runtime import logs
import numpy as np
import sys


def log_probability_function(p):
    r = zeus_pipeline.run_results(p)
    return r.post, (r.prior, r.extra)


class ZeusSampler(ParallelSampler):
    parallel_output = False
    supports_resume = True
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global zeus_pipeline
        zeus_pipeline = self.pipeline

        if self.is_master():
            import emcee
            import zeus
            self.zeus = zeus

            if not hasattr(zeus, "EnsembleSampler"):
                raise ImportError("There are two python packages called Zeus, and you"
                    " have the wrong one.  Uninstall zeus and install zeus-mcmc.")

            # Parameters of the zeus sampler
            self.nwalkers = self.read_ini("walkers", int)
            self.samples = self.read_ini("samples", int)
            self.nsteps = self.read_ini("nsteps", int, 50)

            assert self.nsteps>0, "You specified nsteps<=0 in the ini file - please set a positive integer"
            assert self.samples>0, "You specified samples<=0 in the ini file - please set a positive integer"

            # Parameters controlling initialization
            random_start = self.read_ini("random_start", bool, False)
            start_file = self.read_ini("start_points", str, "")
            covmat_file = self.read_ini("covmat", str, "")
            self.ndim = len(self.pipeline.varied_params)

            # Zeus move options
            move_names = {
                "differential": zeus.moves.DifferentialMove,
                "gaussian": zeus.moves.GaussianMove,
                "global": zeus.moves.GlobalMove,
                "random": zeus.moves.RandomMove,
                "kde": zeus.moves.KDEMove,
            }

            moves = self.read_ini("moves", str, "differential").strip()

            # parse the moves into the list self.moves.
            # Expected format is, e.g. "differential:1.0  global:0.5"
            # but the weights default to 1
            self.moves = []
            moves = moves.split()
            for move in moves:
                # split if weighting present
                if ':' in move:
                    move_name, move_weight = move.split(':')
                else:
                    move_name = move
                    move_weight = 1.0
                move_cls = move_names[move_name]
                self.moves.append((move_cls(), move_weight))

            logs.overview("Running zeus with moves:")
            logs.overview(str(self.moves))

            # Other zeus options
            self.tune = self.read_ini("tune", bool, True)
            self.tolerance = self.read_ini("tolerance", float, 0.05)
            self.maxsteps = self.read_ini("maxsteps", int, 10000)
            self.patience = self.read_ini("patience", int, 5)
            self.maxiter = self.read_ini("maxiter", int, 10000)
            self.verbose = self.read_ini("verbose", bool, False)

            #Starting positions and values for the chain
            self.num_samples = 0
            self.prob0 = None
            self.blob0 = None

            # Generate starting point
            if start_file:
                self.p0 = self.load_start(start_file)
                logs.overview(f"Loaded starting position from {start_file}")
            elif self.distribution_hints.has_cov():
                center = self.start_estimate()
                cov = self.distribution_hints.get_cov()
                self.p0 = sample_ellipsoid(center, cov, size=self.nwalkers)
                logs.overview("Generating starting positions from covmat from earlier in pipeline")
            elif covmat_file:
                center = self.start_estimate()
                cov = self.load_covmat(covmat_file)
                logs.overview(f"Generating starting position from covmat in  {covmat_file}")
                iterations_limit = 100000
                n=0
                p0 = []
                for i in range(iterations_limit):
                    p = sample_ellipsoid(center, cov)[0]
                    if np.isfinite(self.pipeline.prior(p)):
                        p0.append(p)
                    if len(p0)==self.nwalkers:
                        break
                else:
                    raise ValueError("The covmat you used could not generate points inside the prior")
                self.p0 = np.array(p0)
            elif random_start:
                self.p0 = [self.pipeline.randomized_start()
                           for i in range(self.nwalkers)]
                logs.overview("Generating random starting positions from within prior")
            else:
                center_norm = self.pipeline.normalize_vector(self.start_estimate())
                sigma_norm=np.repeat(1e-3, center_norm.size)
                p0_norm = sample_ball(center_norm, sigma_norm, size=self.nwalkers)
                p0_norm[p0_norm<=0] = 0.001
                p0_norm[p0_norm>=1] = 0.999
                self.p0 = [self.pipeline.denormalize_vector(p0_norm_i) for p0_norm_i in p0_norm]
                logs.overview("Generating starting positions in small ball around starting point")

            #Finally we can create the sampler
            self.started = False
            self.sampler = self.zeus.EnsembleSampler(self.nwalkers, self.ndim,
                                                     log_probability_function,
                                                     tune=self.tune,
                                                     tolerance=self.tolerance,
                                                     maxsteps=self.maxsteps,
                                                     patience=self.patience,
                                                     maxiter=self.maxiter,
                                                     verbose=self.verbose,
                                                     pool=self.pool)

    def resume(self):
        resume_info = self.read_resume_info()

        if resume_info is None:
            return

        moves, weights, tune, mu, mus = resume_info

        changed = False
        old_classes = [m.__class__.__name__ for m in moves]
        if len(self.sampler._moves) != len(moves):
            changed = True
        else:
            for m, c in zip(self.sampler._moves, old_classes):
                if m.__class__.__name__ != c:
                    changed = True

        if changed:
            raise ValueError("You have changed the number or type of "
                "the Move objects since re-running this chain. "
                "You can't directly resume a chain with different moves, "
                "so either: (a) change them back (to {}), "
                "(b) delete the sample_status output file, "
                "(c) switch off runtime.resume.  If you just want to restart the "
                "walkers from old positions, set zeus.start_points")

        logs.overview("Loaded 'Move' objects from resume file:")
        for m, w in zip(moves, weights):
            logs.overview(    "{} mu0={}  weight={}  tuning={}".format(m.__class__.__name__, m.mu0, w, m.tune))
        self.sampler._moves = moves
        self.sampler._weights = weights
        self.sampler.tune = tune
        self.sampler.mu = mu
        self.sampler.mus = mus

        if tune:
            logs.overview("Resumed sampler is still tuning")
        else:
            logs.overview("Resumed sampler has finished tuning")

        # if we have some chain, read it here and use it to get
        # the p0 value and number of samples
        if self.output.resumed:
            data = np.genfromtxt(self.output._filename, invalid_raise=False)[:, :self.ndim]
            num_samples = len(data) // self.nwalkers
            self.p0 = data[-self.nwalkers:]
            self.num_samples += num_samples
            if self.num_samples >= self.samples:
                logs.error("You told me to resume the chain - it has already completed (with {} samples), so sampling will end.".format(len(data)))
                logs.error("Increase the 'samples' parameter to keep going.")
            else:
                logs.overview("Continuing zeus from existing chain - have {} samples already".format(len(data)))

    def load_start(self, filename):
        #Load the data and cut to the bits we need.
        #This means you can either just use a test file with
        #starting points, or an emcee output file.
        data = np.genfromtxt(filename, invalid_raise=False)[-self.nwalkers:, :self.ndim]
        if data.shape != (self.nwalkers, self.ndim):
            raise RuntimeError("There are not enough lines or columns "
                               "in the starting point file %s" % filename)
        return list(data)


    def load_covmat(self, covmat_file):
        covmat = np.loadtxt(covmat_file)

        # Convert a single number to a 1x1 array,
        # although it is ambiguous
        if covmat.ndim == 0:
            covmat = covmat.reshape((1, 1))

        # Convert an array of std devs to a covmat
        elif covmat.ndim == 1:
            covmat = np.diag(covmat ** 2)

        # Check the size is right.
        nparams = len(self.pipeline.varied_params)
        if covmat.shape != (nparams, nparams):
            raise ValueError("The covariance matrix was shape (%d x %d), "
                    "but there are %d varied parameters." %
                    (covmat.shape[0], covmat.shape[1], nparams))
        return covmat


    def execute(self):
        #Run the zeus sampler.
        if self.started:
            start = self.sampler.chain.shape[0]
        else:
            logs.error("Begun sampling")
            self.started = True
            start = 0

        # Main execution
        self.sampler.run(self.p0, self.nsteps)
        # Update patience (relevant when self.nsteps<self.patience)
        self.sampler.patience-= self.nsteps
        # record ending point of this iteration
        end = self.sampler.chain.shape[0]

        post = self.sampler.get_log_prob()
        chain = self.sampler.get_chain()
        blobs = self.sampler.get_blobs()

        # Output results per sampler per walker
        for i in range(start, end):
            for j in range(self.nwalkers):
                prior, extra = blobs[i, j]
                self.output.parameters(chain[i, j], extra, prior, post[i, j])

        #Set the starting positions for the next chunk of samples
        #to the last ones for this chunk
        # get_last_sample changed from an attribute to a method recently
        try:
            self.p0 = self.sampler.get_last_sample()
        except TypeError:
            self.p0 = self.sampler.get_last_sample
        self.num_samples += self.nsteps

        taus = self.zeus.AutoCorrTime(chain)
        logs.overview(f"Have {self.num_samples} samples from zeus. Current auto-correlation estimates are:")
        for par, tau in zip(self.pipeline.varied_params, taus):
            logs.overview("   {}:  {:.2f}".format(par, tau))
        sys.stdout.flush()

        if self.sampler.tune:
            logs.overview("Sampler is (still) tuning")
        else:
            logs.overview("Sampler is no longer tuning")

        self.write_resume_info([self.sampler._moves, self.sampler._weights,
                                self.sampler.tune, self.sampler.mu, self.sampler.mus])

    def is_converged(self):
        return self.num_samples >= self.samples




from .. import ParallelSampler
from ...runtime import logs
import numpy as np


def task(p):
    #pipeline here refers to a global variable
    #created later.  This is the only way we know
    #to get MPI task sharing to work
    results = pipeline.run_results(p)
    return (results.prior, results.post, results.extra)


class GridmaxSampler(ParallelSampler):
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline

        if self.is_master():
            self.nsteps = self.read_ini("nsteps", int, 24)
            self.tolerance = self.read_ini("tolerance", float, 0.1)
            self.max_iterations = self.read_ini("max_iterations", int, 10000)
            self.output_ini = self.read_ini("output_ini", str, "")
            self.ndim = len(self.pipeline.varied_params)
            self.p = self.pipeline.normalize_vector(self.pipeline.start_vector())
            self.dimension = 0
            self.bounds = [(0,1) for i in range(self.ndim)]
            self.previous_maxlike = -np.inf
            self.maxlike = -np.inf
            self.iterations = 0


    def execute(self):
        #Do a slice maximization through the current
        #dimension, with the current bounds, in parallel
        #Start by setting all the points to the current one

        normed_points = [self.p.copy() for i in range(self.nsteps)]

        #Figure out which points to sample at in the current 
        #dimension
        d = self.dimension
        bounds = self.bounds[d]
        samples = np.linspace(bounds[0], bounds[1], self.nsteps)

        logs.overview("Minimizing in %s"%self.pipeline.varied_params[d].name)

        #Fill in the sample points for the current
        #dimension in the samples
        for (p,s) in zip(normed_points,samples):
            p[d] = s

        #Denormalize the points to get the physical parameters
        points = [self.pipeline.denormalize_vector(p) for p in normed_points]

        #And send them off to the workers
        if self.pool:
            results = self.pool.map(task, points)
        else:
            results = list(map(task, points))

        #Log the results for posterity
        for p, (pr, po, e) in zip(points, results):
            self.output.parameters(p, e, pr, po)

        #And now update our information.
        #We need to find the two points either side
        #of the maximum likelihood and set them as our new bound,
        #and set our new starting point to be the max value
        posteriors = np.array([r[1] for r in results])
        best = posteriors.argmax()
        #first, the special cases one of the boundary 
        #points is the best (this might be cause for concern, of course)
        if best==0:
            #If we are up against the lower boundary then
            #use the edge as the lower bound, the first point 
            #as the upper, and the half-way point as the new start
            low = 0.0
            high = normed_points[1][d]
            start = high/2.0
        elif best==self.nsteps-1:
            #if on the upper edge, do the mirror image
            low = normed_points[-1][d]
            high = 1.0
            start = (normed_points[-1][d] + 1.0) / 2.0
        else:
            #but the usual case is to just bracket the
            #best-fit point
            low = normed_points[best-1][d]
            high = normed_points[best+1][d]
            start = normed_points[best][d]

        #update the bounds and start with whatever we found
        self.p[d] = start
        self.bounds[d] = (low, high)
    
        #If we have been around all the parameters 
        #then update the best-fits
        if d==0:
            self.previous_maxlike = self.maxlike
            self.maxlike = results[best][0]

        logs.overview(f"New best fit L = {posteriors.max()} at {self.pipeline.varied_params[d].name} = {points[best][d]}")

        #and go on to the next dimension
        self.dimension = (d+1)%self.ndim

        if self.is_converged() and self.output_ini:
            self.pipeline.create_ini(points[best], self.output_ini)
        self.iterations += 1



    def is_converged(self):
        return (self.dimension==1 and (0<self.maxlike-self.previous_maxlike<self.tolerance)) or self.iterations > self.max_iterations

# legay alias
GridMaxSampler = GridmaxSampler



import yaml
import collections
import os

#  Tasks:
# when we run a pipeline, collect all the attribution
# information for it.  Collate into one object.
# Provide a series of comment  lines

class ModuleAttribution(object):

	def __init__(self, data):
		self.data = data
		self.name = data.get('name', 'Unknown')
		self.version = data.get('version', '')
		self.attribution = self.to_list(
			data.get("attribution", "Unknown"))
		self.cite = self.to_list(
			data.get("cite", []))
	@staticmethod
	def to_list(obj):
		if isinstance(obj, list): return obj
		return [obj]

	@classmethod
	def from_yaml(cls,filename):
		if os.path.exists(filename):
			data = yaml.safe_load(open(filename))
		else:
			data = {}
		return cls(data)


class PipelineAttribution(object):
	def __init__(self, modules):
		#for each module, find the associated YAML file,
		#if there is one
		self.attributions = collections.OrderedDict()

		for module in modules:
			directory, _ = os.path.split(module.filename)
			filename = os.path.join(directory, "module.yaml")
			self.attributions[module.name] = ModuleAttribution.from_yaml(filename)

	def write_output(self, output):
		comments = []
		for i,(name, info) in enumerate(self.attributions.items()):
			comment = ""
			out_name = info.name
			if info.name=='Unknown':
				out_name = name
				comment = ("name from ini file")
			if info.version:
				out_name += ' ' + str(info.version)
			output.metadata("module_%d"%i, out_name, comment)
			for c in info.cite:
				output.comment("CITE "+c)





#coding: utf-8
from cosmosis.runtime.parameter import Parameter
from .elements import PostProcessorElement, MCMCPostProcessorElement, WeightedMCMCPostProcessorElement, MultinestPostProcessorElement
try: 
    from getdist import MCSamples
except: 
    MCSamples = None
# getdist not implemented for polychord and grid sampler yet
import numpy as np
import scipy as sp
from io import StringIO
from .utils import std_weight, mean_weight, median_weight, percentile_weight, find_asymmetric_errorbars
from .outputs import PostprocessText, PostprocessTable, MiniTable



class Statistics(PostProcessorElement):
    def __init__(self, *args, **kwargs):
        super(Statistics, self).__init__(*args, **kwargs)

    def run(self):
        print("I do not know how to generate statistics for this kind of data")
        return []
 
    def filename(self, base):
        filename = super(Statistics, self).filename("txt", base)
        return filename

    def get_table_output(self, base, cols):
        filename = self.filename(base)
        f = self.get_output(base)
        if f is None:
            t = MiniTable(cols)
            self.set_output(base, PostprocessTable(base, filename, t))
        else:
            t = f.value
        return t

    def get_text_output(self, base, header="", section_name=""):
        filename = self.filename(base)
        f = self.get_output(base)
        if f is None:
            f = StringIO()
            self.set_output(base, PostprocessText(base,filename,f))
            new_file = True
            if header:
                f.write(header+'\n')
        else:
            f=f.value
            new_file = False
        if section_name:
            f.write("#%s\n"%section_name)
        return f, filename, new_file

class ConstrainingStatistics(Statistics):

    def report_file(self):
        #Get the filenames to make
        return [
            self.report_centroid("means", self.mu),
            self.report_centroid("medians", self.median),
            self.report_limit("l95", self.l95),
            self.report_limit("u95", self.u95),
            self.report_limit("l68", self.l68),
            self.report_limit("u68", self.u68),
            self.report_limit("lerr68", self.lerr68),
            self.report_limit("uerr68", self.uerr68),
            self.report_limit("lerr95", self.lerr95),
            self.report_limit("uerr95", self.uerr95),
            self.report_limit("peak1d", self.peak1d),
            self.report_max_post("maxpost", self.maxpost),
            self.report_max_post("maxlike", self.maxlike),
        ]
    
    def report_max_post(self, kind, values):
        cols = ["parameter", kind, "data_set"]
        table = self.get_table_output(kind, cols)
        for (col, x) in zip(self.source.colnames, values):
            table.append([col, x, self.source.label])
        return table

    def report_centroid(self, kind, col):
        cols = ["parameter", kind, "sigma", "data_set"]
        table = self.get_table_output(kind, cols)
        for (col, x, sigma) in zip(self.source.colnames, col, self.sigma):
            table.append([col, x, sigma, self.source.label])
        return table

    def report_limit(self, kind, col):
        cols = ["parameter", kind, "data_set"]
        table = self.get_table_output(kind, cols)
        for (col, x) in zip(self.source.colnames, col):
            table.append([col, x, self.source.label])
        return table

    @staticmethod
    def find_median(x, P):
        C = [0] + P.cumsum()
        return np.interp(C[-1]/2.0,C,x)

    @staticmethod
    def find_percentile(x, P, p):
        C = [0] + P.cumsum()
        return np.interp(C[-1]*p/100.0,C,x)

    def report_screen(self):
        self.report_screen_mean()
        self.report_screen_asym()
        self.report_screen_asym95()
        self.report_screen_median()
        self.report_screen_limits()
        self.report_screen_maxpost()
        self.report_screen_maxlike()
        #Print the same summary stats that go into the
        #files but to the screen instead, in a pretty format        

    def report_screen_maxpost(self):
        print()
        print("Maximum posterior:")
        for P in zip(self.source.colnames, self.maxpost):
            print('    %s = %g' % P)
        print()

    def report_screen_maxlike(self):
        print()
        print("Maximum likelihood:")
        for P in zip(self.source.colnames, self.maxlike):
            print('    %s = %g' % P)
        print()

    def report_screen_mean(self):
        #Means
        print()
        print("Marginalized mean, std-dev:")
        for P in zip(self.source.colnames, self.mu, self.sigma):
            print('    %s = %g ± %g ' % P)
        print()

    def report_screen_asym(self):
        #Means
        print()
        print("Marginalized 1D peak, 68% asymmetric error bars:")
        for P in zip(self.source.colnames, self.peak1d, self.lerr68, self.uerr68):
            name,mu,lerr,uerr = P
            print('    %s = %g + %g - %g ' % (name,mu,uerr-mu, mu-lerr))
        print()

    def report_screen_asym95(self):
        #Means
        print()
        print("Marginalized 1D peak, 95% asymmetric error bars:")
        for P in zip(self.source.colnames, self.peak1d, self.lerr95, self.uerr95):
            name,mu,lerr,uerr = P
            print('    %s = %g + %g - %g ' % (name,mu,uerr-mu, mu-lerr))
        print()

    def report_screen_median(self):
        #Medians
        print("Marginalized median, std-dev:")
        for P in zip(self.source.colnames, self.median, self.sigma):
            print('    %s = %g ± %g' % P)
        print()

    def report_screen_limits(self):
        #Mode
        print("95% lower limits:")
        for name, val in zip(self.source.colnames, self.l95):
            print('    %s > %g' % (name, val))
        print()
        print("95% upper limits:")
        for name, val in zip(self.source.colnames, self.u95):
            print('    %s < %g' % (name, val))
        print()

        #Mode
        print("68% lower limits:")
        for name, val in zip(self.source.colnames, self.l68):
            print('    %s > %g' % (name, val))
        print()
        print("68% upper limits:")
        for name, val in zip(self.source.colnames, self.u68):
            print('    %s < %g' % (name, val))
        print()

    @staticmethod
    def likelihood_ratio_warning(marge_like, name):
        #Check for an warn about a bad likelihood ratio,
        #which would indicate that the likelihood did not fall
        #off by the edges        
        if marge_like.min()==0: return

        like_ratio = marge_like.max() / marge_like.min()
        if like_ratio < 20:
            print()
            print("L_max/L_min = %f for %s." % (like_ratio, name))
            print("This indicates that the grid did not go far from the peak in this dimension")
            print("Marginalized values will definitely be poor estimates for this parameter, and probably")
            print("for any other parameters correlated with this one")


class MetropolisHastingsStatistics(ConstrainingStatistics, MCMCPostProcessorElement):
    def compute_basic_stats_col(self, col):
        data = self.reduced_col(col)
        n = len(data)
        try:
            peak1d, ((lerr68, uerr68), (lerr95, uerr95)) = find_asymmetric_errorbars([0.68, 0.95], data)
        except (RuntimeError, ValueError, sp.linalg.LinAlgError):
            (lerr68, uerr68), (lerr95, uerr95) = (np.nan, np.nan), (np.nan, np.nan)
            peak1d = np.nan
        post = self.reduced_col("post")
        try:
            like = self.reduced_col("like")
        except ValueError:
            like = post - self.reduced_col("prior")
        maxpost = data[post.argmax()]
        maxlike = data[like.argmax()]
        return n, data.mean(), data.std(), np.median(data), np.percentile(data, 32.), np.percentile(data, 68.), np.percentile(data, 5.), np.percentile(data, 95.), lerr68, uerr68, lerr95, uerr95, peak1d, maxlike, maxpost

    def compute_basic_stats(self):
        self.mu = []
        self.sigma = []
        self.median = []
        self.l68 = []
        self.u68 = []
        self.l95 = []
        self.u95 = []
        self.lerr68 = []
        self.uerr68 = []
        self.lerr95 = []
        self.uerr95 = []
        self.peak1d = []
        self.maxpost = []
        self.maxlike = []
        try:
            self.best_fit_index = self.source.get_col("post").argmax()
        except:
            self.best_fit_index = self.source.get_col("like").argmax()
        
        n = 0
        for col in self.source.colnames:
            n, mu, sigma, median, l68, u68, l95, u95, lerr68, uerr68, lerr95, uerr95, peak1d, maxlike, maxpost = self.compute_basic_stats_col(col)
            self.mu.append(mu)
            self.sigma.append(sigma)
            self.median.append(median)
            self.l68.append(l68)
            self.u68.append(u68)
            self.l95.append(l95)
            self.u95.append(u95)
            self.lerr68.append(lerr68)
            self.uerr68.append(uerr68)
            self.lerr95.append(lerr95)
            self.uerr95.append(uerr95)
            self.peak1d.append(peak1d)
            self.maxlike.append(maxlike)
            self.maxpost.append(maxpost)
        return n

    def get_gdobj(self):
        datapts = []
        for col in self.source.colnames:
            datapts.append(self.reduced_col(col))
        datapts = np.array(datapts).T
        vlfile = self.source.extract_ini("VALUES")#test
        vlpars = Parameter.load_parameters(vlfile)#test
        rangedict = {}
        for vlpar in vlpars:
            rangedict[str(vlpar)] = np.array(vlpar.limits)
            
        try: 
            loglikes = self.source.reduced_col("post")
        except: 
            loglikes = self.source.reduced_col("like")

        if MCSamples:
            gdc = MCSamples(samples = datapts,loglikes=loglikes,names=self.source.colnames,name_tag = self.source.name,ranges=rangedict)# ranges from value file
        else:
            raise ImportError('GetDist is not installed')
        self.source.gdc = gdc
        return gdc

    def compute_basic_statsgd_col(self, gdc, col):
        dens1d = gdc.get1DDensity(col,writeDataToFile=False)
        if dens1d is None:
            return [np.nan for i in range(9)]

        def func(x):
            return -dens1d.Prob(x)

        results = [
            sp.optimize.fmin(func,dens1d.bounds()[0])[0],
            #l/u are done before KDE, lower or upper
            gdc.confidence(col,0.32),
            gdc.confidence(col,0.32,True),
            gdc.confidence(col,0.05),
            gdc.confidence(col,0.05,True),
            #l/uerr are done after KDE
            dens1d.getLimits([0.68])[0],
            dens1d.getLimits([0.68])[1],
            dens1d.getLimits([0.95])[0],
            dens1d.getLimits([0.95])[1],
        ]
        return results


    def compute_basic_statsgd(self):
        self.mu = []
        self.sigma = []
        self.median = []
        self.l68 = []
        self.u68 = []
        self.l95 = []
        self.u95 = []
        self.lerr68 = []
        self.uerr68 = []
        self.lerr95 = []
        self.uerr95 = []
        self.peak1d = []
        self.maxlike = []
        self.maxpost = []
        try:
            self.best_fit_index = self.source.get_col("post").argmax()
        except:
            self.best_fit_index = self.source.get_col("like").argmax()
        n = 0
        gdc = self.get_gdobj()
        # output: self.mu, sigma, median, l/u68, l/u95, l/uerr68, l/uerr95

        n = len(gdc.samples)
        self.mu = gdc.getMeans()
        self.sigma = np.sqrt(gdc.cov().diagonal())
        self.median = np.median(gdc.samples,axis=0)

        for col in self.source.colnames:
            col_results = self.compute_basic_statsgd_col(gdc, col)
            self.peak1d.append(col_results[0])
            self.l68.append(col_results[1])
            self.u68.append(col_results[2])
            self.l95.append(col_results[3])
            self.u95.append(col_results[4])
            self.lerr68.append(col_results[5])
            self.uerr68.append(col_results[6])
            self.lerr95.append(col_results[7])
            self.uerr95.append(col_results[8])
            self.uerr95.append(col_results[9])
            self.uerr95.append(col_results[10])
        return n



    def run(self):
        if self.source.options.get("getdist",False):
            N = self.compute_basic_statsgd()
        else:
            N = self.compute_basic_stats()
        print("Samples after cutting:", N)

        self.report_screen()
        files = self.report_file()
        return files

class ChainCovariance(object):
    def run(self):
        #Determine the parameters to use

        col_names = [p for p in self.source.colnames if p.lower() not in ["like","post", "importance", "weight", "prior"]]

        if len(col_names)<2:
            return []

        if self.source.options.get("getdist",False):
            # Have to manually skip the first bit here
            covmat = self.source.gdc.cov()[:len(col_names),:len(col_names)]
        else:

            if hasattr(self, 'weight_col'):
                w = self.weight_col()
            else:
                w = None

            cols = [self.reduced_col(p) for p in col_names]
            covmat = np.cov(cols, aweights=w)

        #For the proposal we just want the first 
        #nvaried rows/cols - we don't want to include
        #extra parameters like sigma8
        n = self.source.metadata[0]['n_varied']
        proposal = covmat[:n,:n]

        #Save the covariance matrix
        t1 = self.get_table_output("covmat", col_names)
        if len(t1) == 0:
            for row in covmat:
                t1.append(row)
        else:
            print("NOT saving more than one covariance matrix - just using first ini file")

        #Save the proposal matrix
        t2 = self.get_table_output("proposal", col_names[:n])
        if len(t2) == 0:
            for row in proposal:
                t2.append(row)
        else:
            print("NOT saving more than one proposal matrix - just using first ini file")
        return [t1, t2]

class MetropolisHastingsCovariance(ChainCovariance, Statistics, MCMCPostProcessorElement):
    pass

class WeightedMetropolisHastingsCovariance(ChainCovariance, Statistics, MCMCPostProcessorElement):
    pass

class GridStatistics(ConstrainingStatistics):
    def set_data(self):
        self.nsample = int(self.source.sampler_option("nsample_dimension"))
        self.nrow = len(self.source)
        self.ncol = int(self.source.sampler_option('n_varied'))

        extra = self.source.sampler_option("extra_output","").replace('/','--').split()
        self.grid_columns = [i for i in range(self.ncol) if (not self.source.colnames[i] in extra) and (self.source.colnames[i]!="post") and (self.source.colnames[i]!="like")]
        self.ndim = len(self.grid_columns)
        assert self.nrow == self.nsample**self.ndim
        self.shape = np.repeat(self.nsample, self.ndim)

        try:
            like = self.source.get_col("post").reshape(self.shape).copy()
        except:
            like = self.source.get_col("like").reshape(self.shape).copy()

        like -= like.max()

        self.like = np.exp(like).reshape(self.shape)

        grid_names = [self.source.colnames[i] for i in range(self.ncol) if i in self.grid_columns]
        self.grid = [np.unique(self.source.get_col(name)) for name in grid_names]

    def run(self):
        self.set_data()
        self.compute_stats()
        self.report_screen()
        files = self.report_file()
        return files

    def report_screen(self):
        self.report_screen_mean()
        #self.report_screen_asym()
        # self.report_screen_asym95()
        self.report_screen_median()
        #self.report_screen_limits()

    def report_file(self):
        #Get the filenames to make
        return [
            self.report_centroid("means", self.mu),
            self.report_centroid("medians", self.median),
            # self.report_centroid("modes", self.peak1d),
            self.report_limit("l95", self.l95),
            self.report_limit("u95", self.u95),
            ]

    def compute_stats(self):
        #1D stats
        self.mu = np.zeros(self.ncol)
        self.median = np.zeros(self.ncol)
        self.sigma = np.zeros(self.ncol)
        self.l95 = np.zeros(self.ncol)
        self.u95 = np.zeros(self.ncol)        
        try:like = self.source.get_col("post")
        except:like = self.source.get_col("like")
        self.best_fit_index = np.argmax(like)
        #Loop through colums
        for i, name in enumerate(self.source.colnames[:self.ncol]):
            if i in self.grid_columns:
                self.mu[i], self.median[i], self.sigma[i], self.l95[i], self.u95[i] = self.compute_grid_stats(i)
            else:
                self.mu[i], self.median[i], self.sigma[i], self.l95[i], self.u95[i] = self.compute_derived_stats(i)




    def compute_grid_stats(self, i):
        name = self.source.colnames[i]
        col = self.source.get_col(name)
        #Sum the likelihood over all the axes other than this one
        #to get the marginalized likelihood
        marge_like = self.like.sum(tuple(j for j in range(self.ndim) if j!=i))
        marge_like = marge_like / marge_like.sum()
        
        #Find the grid points with this value
        vals = self.grid[i]

        #A quick potential error warning
        self.likelihood_ratio_warning(marge_like, name)

        #Compute the statistics
        mu = (vals*marge_like).sum()
        sigma2 = ((vals-mu)**2*marge_like).sum()
        median = self.find_median(vals, marge_like)
        l95 = self.find_percentile(vals, marge_like, 5.0)
        u95 = self.find_percentile(vals, marge_like, 95.0)
        return mu, median, sigma2**0.5, l95, u95

    def compute_derived_stats(self, i):
        #This is a bit simpler - just need to 
        #sum over everything
        name = self.source.colnames[i]
        col = self.source.get_col(name)
        try:like = self.source.get_col("post")
        except:like = self.source.get_col("like")
        like = like / like.sum()
        mu = (col*like).sum()
        sigma2 = ((col-mu)**2*like).sum()
        median = self.find_median(col, like)
        l95 = self.find_percentile(col, like, 5.0)
        u95 = self.find_percentile(col, like, 95.0)        
        return mu, median, sigma2**0.5, l95, u95


class TestStatistics(Statistics):
    def run(self):
        return []


class GelmanRubinStatistic(MetropolisHastingsStatistics):
    def gelman_rubin(self, name):
        # This simplified form compared to the online analytics code:
        # - assumes the chains are fairly long
        # - does one parameter at a time

        # Get the chains for each input file
        chains = self.source.reduced_col(name,stacked=False)

        steps = min([len(chain) for chain in chains])
        chains = [chain[:steps] for chain in chains]
        means = [chain.mean() for chain in chains]
        variances = [chain.var() for chain in chains]

        number_chains = len(chains)

        B_over_n = np.var(means, ddof=1)
        B = B_over_n * steps
        W = np.mean(variances)
        V = W + (1. + 1./number_chains) * B_over_n
        # TODO: check for 0-values in W
        Rhat = np.sqrt(V/W)
        return Rhat - 1.0

    def run(self):
        if len(self.source.data)<2:
            print()
            print("(One chain found. Run multiple chains if you want the Gelman-Rubin test)")
            print()
            return []
        names = [c for c in self.source.colnames if  c not in ['weight', 'like', 'post']]
        cols = ["parameter", "R-1", "data_set"]
        t = self.get_table_output("gelman", cols)

        print()
        print("Gelman-Rubin tests")
        print("------------------")
        print("(Variance of means / Mean of variances.  Smaller is better, a few percent is usually good convergence)")
        print()
        for name in names:
            R1 = self.gelman_rubin(name)
            t.append([name, R1, self.source.label])

            if R1>0.1:
                print("{}    {}  -- POORLY CONVERGED PARAMETER AT 10% LEVEL".format(name,R1))
            else:
                print("{}    {}".format(name,R1))
        print()
        return [t]




class DunkleyTest(MetropolisHastingsStatistics):
    """
    Run the Dunley et al (2005) power spectrum test
    for MCMC convergence.  This is, loosely speaking,
    the Fourier domain version of an auto-correlation
    length check.

    """
    #This is the value recommended in Dunkely et al.
    #You can probably get away with less than this
    jstar_convergence_limit = 20.0

    def run(self):
        n = self.source.metadata[0]['n_varied']
        jstar = []
        #Just sample over the sample parameters,
        #not the likelihood or the derived parameters
        params = self.source.colnames[:n]
        print("Dunkely et al (2005) power spectrum test.")
        print("For converged chains j* > %.1f:" %self.jstar_convergence_limit)

        cols = ["parameter", "jstar", "data_set"]
        t = self.get_table_output("dunkley", cols)


        for param in params:
            cols = self.reduced_col(param, stacked=False)
            for c,col in enumerate(cols):
                js = self.compute_jstar(col)
                jstar.append(js)
                m = "Chain %d:  %-35s j* = %-.1f" % (c+1, param, js)
                if js>20:
                    print("    %-50s" % m)
                else:
                    print("    %-50s NOT CONVERGED!" % m)
                
                t.append([param, js, self.source.label])
        print()
        if not np.min(jstar)>self.jstar_convergence_limit:
            print("The Dunkley et al (2005) power spectrum test shows that this chain has NOT CONVERGED.")
            print("It is quite a conservative test, so no need to panic.")
        else:
            print("The power spectra for this chain suggests good convergence.")
        print()
        return [t]


    @staticmethod
    def compute_jstar(x):
        import scipy.optimize

        #Get the power spectrum of the chain
        n=len(x)
        p = abs(np.fft.rfft(x)[1:n//2])**2
        #And the k-axis
        j = np.arange(p.size)+1.
        k = j / (2*np.pi*n)
        #fitting is done on the log of the power 
        #spectrum
        logp = np.log(p)

        #The model for the power spectrum.
        #See Dunkley et al for info on the 
        #constant
        def template(k, L0, a, kstar):
            K = (kstar / k)**a
            euler_mascheroni = 0.5772156649015
            return L0 + np.log(K/(1.+K)) - euler_mascheroni

        #Starting guess values for parameters.
        #These are usually fine.
        L0 = logp[0:10].mean()
        a  = 1.0
        kstar = 0.1/n

        #Fit curve with LevMar least-squares
        start_params = [L0, a, kstar]
        try:
            params, cov = scipy.optimize.curve_fit(template, k, logp, start_params)
        except RuntimeError:
            #Runtime Errors signal that the fitting process has failed.
            #This usually happens because the chain is too short,
            #or has periodic or other features in like an emcee chain
            params = [np.nan, np.nan, np.nan]
        L0, a, kstar = params
        return kstar * x.size * 2 * np.pi


class WeightedStatistics(object):
    def get_gdobj(self):
        datapts = []
        weight = self.weight_col()
        for col in self.source.colnames:
            datapts.append(self.reduced_col(col))
        datapts = np.array(datapts).T
        vlfile = self.source.extract_ini("VALUES")#test
        vlpars = Parameter.load_parameters(vlfile)#test
        rangedict = {}
        for vlpar in vlpars:
            rangedict[str(vlpar)] = np.array(vlpar.limits)
        try: 
            loglikes = self.source.reduced_col("post")
        except: 
            loglikes = self.source.reduced_col("like")

        if MCSamples:
            gdc = MCSamples(samples = datapts,weights=weight,loglikes=loglikes, names=self.source.colnames,name_tag = self.source.label,ranges=rangedict)# ranges from value file
        else:
            raise ImportError('GetDist is not installed')
        self.source.gdc = gdc
        return gdc
        
    def compute_basic_stats_col(self, col):
        data = self.reduced_col(col)
        weight = self.weight_col()
        n = len(data)
        try:
            print(col)
            peak1d, ((lerr68, uerr68), (lerr95, uerr95)) = find_asymmetric_errorbars([0.68, 0.95], data, weight)
        except (RuntimeError, ValueError, sp.linalg.LinAlgError):
            (lerr68, uerr68), (lerr95, uerr95) = (np.nan, np.nan), (np.nan, np.nan)
            peak1d = np.nan
        
        post = self.reduced_col("post")
        try:
            like = self.reduced_col("like")
        except ValueError:
            like = post - self.reduced_col("prior")
        maxpost = data[post.argmax()]
        maxlike = data[like.argmax()]


        return (n, mean_weight(data,weight), std_weight(data,weight), 
            median_weight(data, weight), percentile_weight(data, weight, 32.), percentile_weight(data, weight, 68.),
            percentile_weight(data, weight, 5.), percentile_weight(data, weight, 95.), lerr68, uerr68, lerr95, uerr95, peak1d,
            maxlike, maxpost)


class MultinestStatistics(WeightedStatistics, MultinestPostProcessorElement, MetropolisHastingsStatistics):
    def run(self):
        # Use parent statistics, except add evidence information,
        # which is just read from the file
        files = super(MultinestStatistics,self).run()
        logz = self.source.final_metadata[0]["log_z"]
        try:
            logz_sigma = self.source.final_metadata[0]["log_z_error"]
        except KeyError:
            logz_sigma = np.nan
        #First print to screen
        print("Bayesian evidence:")
        print(f"    log(Z) = {logz:.2f} ± {logz_sigma}")
        print()


        weight = self.weight_col()
        w = weight/weight.max()
        n_eff = w.sum()

        print("Effective number samples = ", n_eff)
        print()
        #Now save to file
        cols=["logz", "logz_sigma", "data_set"]
        t = self.get_table_output("evidence", cols)
        t.append([logz, logz_sigma, self.source.label])

        #Include evidence in list of created files
        files.append(t)
        return files


class PolychordStatistics(MultinestStatistics):
    pass

#The class hierarchy is getting too complex for this - revise it
class WeightedMetropolisStatistics(WeightedStatistics, ConstrainingStatistics, WeightedMCMCPostProcessorElement):
    def compute_basic_statsgd_col(self, gdc, col):
        dens1d = gdc.get1DDensity(col,writeDataToFile=False)
        if dens1d is None:
            return [np.nan for i in range(11)]
        
        like = self.reduced_col("like")
        post = self.reduced_col("post")
        data1 = self.reduced_col(col)
        maxlike = data1[like.argmax()]
        maxpost = data1[post.argmax()]

        def func(x):
            return -dens1d.Prob(x)

        results = [
            sp.optimize.fmin(func,dens1d.bounds()[0])[0],
            #l/u are done before KDE, lower or upper
            gdc.confidence(col,0.32),
            gdc.confidence(col,0.32,True),
            gdc.confidence(col,0.05),
            gdc.confidence(col,0.05,True),
            #l/uerr are done after KDE
            dens1d.getLimits([0.68])[0],
            dens1d.getLimits([0.68])[1],
            dens1d.getLimits([0.95])[0],
            dens1d.getLimits([0.95])[1],
            maxlike,
            maxpost
        ]
        return results


    def compute_basic_statsgd(self):
        self.mu = []
        self.sigma = []
        self.median = []
        self.l68 = []
        self.u68 = []
        self.l95 = []
        self.u95 = []
        self.lerr68 = []
        self.uerr68 = []
        self.lerr95 = []
        self.uerr95 = []
        self.peak1d = []
        self.maxlike = []
        self.maxpost = []
        try:
            self.best_fit_index = self.source.get_col("post").argmax()
        except:
            self.best_fit_index = self.source.get_col("like").argmax()
        n = 0
        gdc = self.get_gdobj()
        # output: self.mu, sigma, median, l/u68, l/u95, l/uerr68, l/uerr95

        n = len(gdc.samples)
        self.mu = gdc.getMeans()
        self.sigma = np.sqrt(gdc.cov().diagonal())
        self.median = np.median(gdc.samples,axis=0)

        for col in self.source.colnames:
            col_results = self.compute_basic_statsgd_col(gdc, col)
            self.peak1d.append(col_results[0])
            self.l68.append(col_results[1])
            self.u68.append(col_results[2])
            self.l95.append(col_results[3])
            self.u95.append(col_results[4])
            self.lerr68.append(col_results[5])
            self.uerr68.append(col_results[6])
            self.lerr95.append(col_results[7])
            self.uerr95.append(col_results[8])
            self.maxlike.append(col_results[9])
            self.maxpost.append(col_results[10])
        return n

        
    def compute_basic_stats(self):
        self.mu = []
        self.sigma = []
        self.median = []
        self.l68 = []
        self.u68 = []
        self.l95 = []
        self.u95 = []
        self.lerr68 = []
        self.uerr68 = []
        self.lerr95 = []
        self.uerr95 = []
        self.peak1d = []
        self.uerr95 = []
        self.peak1d = []
        self.maxlike = []
        self.maxpost = []
        try:
            self.best_fit_index = self.source.get_col("post").argmax()
        except:
            self.best_fit_index = self.source.get_col("like").argmax()
        
        n = 0
        for col in self.source.colnames:
            n, mu, sigma, median, l68, u68, l95, u95, lerr68, uerr68, lerr95, uerr95, peak1d, maxlike, maxpost = self.compute_basic_stats_col(col)
            self.mu.append(mu)
            self.sigma.append(sigma)
            self.median.append(median)
            self.l68.append(l68)
            self.u68.append(u68)
            self.l95.append(l95)
            self.u95.append(u95)
            self.lerr68.append(lerr68)
            self.uerr68.append(uerr68)
            self.lerr95.append(lerr95)
            self.uerr95.append(uerr95)
            self.peak1d.append(peak1d)
            self.maxlike.append(maxlike)
            self.maxpost.append(maxpost)
        return n
        
    def run(self):

        if self.source.options.get("getdist",False):
            N = self.compute_basic_statsgd()
        else:
            N = self.compute_basic_stats()
        print("Samples after cutting:", N)

        self.report_screen()
        files = self.report_file()
        return files

class MultinestCovariance(ChainCovariance, Statistics, MultinestPostProcessorElement):
    pass

class PolychordCovariance(MultinestCovariance):
    pass


class CovarianceMatrix1D(Statistics):
    def run(self):
        Sigma = np.linalg.inv(self.source.data[0]).diagonal()**0.5
        Mu = [float(self.source.metadata[0]['mu_{0}'.format(i)]) for i in range(Sigma.size)]        
        cols = ['param', 'mean', 'std-dev', 'data_set']
        t = self.get_table_output("means", cols)

        for p, mu, sigma in zip(self.source.colnames, Mu, Sigma):
            t.append([p, mu, sigma, self.source.label])

        print()
        print("Marginalized mean, std-dev:")
        for P in zip(self.source.colnames, Mu, Sigma):
            print('    %s = %g ± %g' % P)
        print()

        return [t]

class CovarianceMatrixEllipseAreas(Statistics):
    def run(self):
        params = self.source.colnames
        cols = ["param1", "param2", "area", "figure_of_merit", "data_set"]
        t = self.get_table_output("ellipse_areas", cols)

        covmat_estimate = np.linalg.inv(self.source.data[0])
        for i,p1 in enumerate(params[:]):
            for j,p2 in enumerate(params[:]):
                if j>=i: continue
                #Get the 2x2 sub-matrix
                C = covmat_estimate[:,[i,j]][[i,j],:]
                area = 6.17 * np.pi * np.sqrt(np.linalg.det(C))
                fom = 1.0/area
                t.append([p1, p2, area, fom, self.source.label])

        return [t]


class FisherFigureOfMerit(Statistics):
    def run(self):
        cols = ["figure_of_merit_reduced", "data_set"]
        t = self.get_table_output("fisher_fom", cols)
        F = self.source.data[0]
        n = self.source.metadata[0]['n_varied']
        fom = (np.linalg.det(F))**(-0.5 / n)
        t.append([fom, self.source.label])
        return [t]


class Citations(Statistics):
    #This isn't really a statistic but it uses all the same
    #mechanisms
    def run(self):
        print() 
        message = "#You should cite these papers in any publication based on this pipeline."
        print(message)
        citations = set()
        f, filename, new_file = self.get_text_output("citations", message, self.source.label)
        for comment_set in self.source.comments:
            for comment in comment_set:
                comment = comment.strip()
                if comment.startswith("CITE"):
                    citation =comment[4:].strip()
                    citations.add(citation)
        for citation in citations:
            print("    ", citation)
            f.write("%s\n"%citation)
        print()
        return [filename]



from ...runtime import logs
from .. import ParallelSampler, sample_ellipsoid, sample_ball
import numpy as np
import sys


def log_probability_function(p):
    r = emcee_pipeline.run_results(p)
    return r.post, (r.prior, r.extra)


class EmceeSampler(ParallelSampler):
    parallel_output = False
    supports_resume = True
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global emcee_pipeline
        emcee_pipeline = self.pipeline

        if self.is_master():
            import emcee
            self.emcee = emcee

            self.emcee_version = int(self.emcee.__version__[0])


            # Parameters of the emcee sampler
            self.nwalkers = self.read_ini("walkers", int, 2)
            self.samples = self.read_ini("samples", int, 1000)
            self.nsteps = self.read_ini("nsteps", int, 100)
            self.a = self.read_ini("a", float, 2.0)

            assert self.nsteps>0, "You specified nsteps<=0 in the ini file - please set a positive integer"
            assert self.samples>0, "You specified samples<=0 in the ini file - please set a positive integer"

            random_start = self.read_ini("random_start", bool, False)
            start_file = self.read_ini("start_points", str, "")
            covmat_file = self.read_ini("covmat", str, "")
            self.ndim = len(self.pipeline.varied_params)

            #Starting positions and values for the chain
            self.num_samples = 0
            self.prob0 = None
            self.blob0 = None

            if start_file:
                self.p0 = self.load_start(start_file)
                logs.overview(f"Loaded starting position from {start_file}")
            elif self.distribution_hints.has_cov():
                center = self.start_estimate()
                cov = self.distribution_hints.get_cov()
                self.p0 = sample_ellipsoid(center, cov, size=self.nwalkers)
                logs.overview("Generating starting positions from covmat from earlier in pipeline")
            elif covmat_file:
                center = self.start_estimate()
                cov = self.load_covmat(covmat_file)
                logs.overview(f"Generating starting position from covmat in  {covmat_file}")
                iterations_limit = 100000
                n=0
                p0 = []
                for i in range(iterations_limit):
                    p = sample_ellipsoid(center, cov)[0]
                    if np.isfinite(self.pipeline.prior(p)):
                        p0.append(p)
                    if len(p0)==self.nwalkers:
                        break
                else:
                    raise ValueError("The covmat you used could not generate points inside the prior")
                self.p0 = np.array(p0)
            elif random_start:
                self.p0 = [self.pipeline.randomized_start()
                           for i in range(self.nwalkers)]
                logs.overview("Generating random starting positions from within prior")
            else:
                center_norm = self.pipeline.normalize_vector(self.start_estimate())
                sigma_norm=np.repeat(1e-3, center_norm.size)
                p0_norm = sample_ball(center_norm, sigma_norm, size=self.nwalkers)
                p0_norm[p0_norm<=0] = 0.001
                p0_norm[p0_norm>=1] = 0.999
                self.p0 = [self.pipeline.denormalize_vector(p0_norm_i) for p0_norm_i in p0_norm]
                logs.overview("Generating starting positions in small ball around starting point")

            if self.emcee_version < 3:
                kw = {"a": self.a}
            else:
                kw = {"moves": [(emcee.moves.StretchMove(a=self.a), 1.0)]}

            #Finally we can create the sampler
            self.ensemble = self.emcee.EnsembleSampler(self.nwalkers, self.ndim,
                                                       log_probability_function,
                                                       pool=self.pool, **kw)

    def resume(self):
        if self.output.resumed:
            data = np.genfromtxt(self.output._filename, invalid_raise=False)[:, :self.ndim]
            num_samples = len(data) // self.nwalkers
            self.p0 = data[-self.nwalkers:]
            self.num_samples += num_samples
            if self.num_samples >= self.samples:
                logs.error("You told me to resume the chain - it has already completed (with {} samples), so sampling will end.".format(len(data)))
                logs.error("Increase the 'samples' parameter to keep going.")
            else:
                logs.overview("Continuing emcee from existing chain - have {} samples already".format(len(data)))

    def load_start(self, filename):
        #Load the data and cut to the bits we need.
        #This means you can either just use a test file with
        #starting points, or an emcee output file.
        data = np.genfromtxt(filename, invalid_raise=False)[-self.nwalkers:, :self.ndim]
        if data.shape != (self.nwalkers, self.ndim):
            raise RuntimeError("There are not enough lines or columns "
                               "in the starting point file %s" % filename)
        return list(data)


    def load_covmat(self, covmat_file):
        covmat = np.loadtxt(covmat_file)

        if covmat.ndim == 0:
            covmat = covmat.reshape((1, 1))
        elif covmat.ndim == 1:
            covmat = np.diag(covmat ** 2)

        nparams = len(self.pipeline.varied_params)
        if covmat.shape != (nparams, nparams):
            raise ValueError("The covariance matrix was shape (%d x %d), "
                    "but there are %d varied parameters." %
                    (covmat.shape[0], covmat.shape[1], nparams))
        return covmat


    def output_samples(self, pos, prob, extra_info):
        for params, post, extra in zip(pos,prob,extra_info):
            prior, extra = extra      
            self.output.parameters(params, extra, prior, post)

    def execute(self):
        #Run the emcee sampler.
        if self.num_samples == 0:
            logs.overview("Begun sampling")
        outputs = []
        if self.emcee_version < 3:
            kwargs = dict(lnprob0=self.prob0, blobs0=self.blob0, 
                          iterations=self.nsteps, storechain=False)
        else:
            # In emcee3 we have to enable storing the chain because
            # we want the acceptance fraction.  Also the name of one
            # of the parameters has changed.
            kwargs = dict(log_prob0=self.prob0, blobs0=self.blob0, 
                          iterations=self.nsteps, store=True)

        for (pos, prob, rstate, extra_info) in self.ensemble.sample(self.p0, **kwargs):
            outputs.append((pos.copy(), prob.copy(), np.copy(extra_info)))
    
        for (pos, prob, extra_info) in outputs:
            self.output_samples(pos, prob, extra_info)

        #Set the starting positions for the next chunk of samples
        #to the last ones for this chunk
        self.p0 = pos
        self.prob0 = prob
        self.blob0 = extra_info
        self.num_samples += self.nsteps
        acceptance_fraction = self.ensemble.acceptance_fraction.mean()
        logs.overview("Done {} iterations of emcee. Acceptance fraction {:.3f}".format(
            self.num_samples, acceptance_fraction))
        sys.stdout.flush()
        self.output.final("mean_acceptance_fraction", acceptance_fraction)

    def is_converged(self):
        return self.num_samples >= self.samples



import itertools
import collections
import numpy as np
from ...runtime import logs

from .. import ParallelSampler


INI_SECTION = "importance"


def task(p):
    r = importance_pipeline.posterior(p)
    return r


class ImportanceSampler(ParallelSampler):
    #there are a bunch of these really, but we 
    #need to add them manually so that we can
    #keep the likelihood column at the end
    #so the postprocessors don't need to be re-written
    sampler_outputs = []
    parallel_output = False

    def config(self):
        global importance_pipeline
        importance_pipeline = self.pipeline
        self.input_filename = self.read_ini("input", str)
        self.nstep = self.read_ini("nstep", int, 128)
        self.add_to_likelihood = self.read_ini("add_to_likelihood", bool, False)

        self.converged = False
        if self.is_master():
            self.load_samples(self.input_filename)

    def load_samples(self, filename):
        options = {"filename":filename}
        col_names, cols, metadata, comments, final_metadata = self.output.__class__.load_from_options(options)
        # pull out the "post" column first
        col_names = [name.lower() for name in col_names]
        likelihood_index = col_names.index('post')
        if likelihood_index<0:
            raise ValueError("I could not find a 'post' column in the chain %s"%filename)
        self.original_likelihoods = cols[0].T[likelihood_index]

        #We split the parameters into three groups:
        #   - ones that we have listed as varying
        #   - one that are fixed - WHAT SHOULD WE DO ABOUT THESE???
        #   - extras ones - these should be saved and put in the output

        self.original_extras = collections.OrderedDict()
        self.samples = []
        self.varied_params = []
        self.number_samples = len(cols[0])
        logs.overview("Have %d samples from old chain." % self.number_samples)
        for code,col in zip(col_names, cols[0].T):
            #we have already handled the likelihood
            if code=='post':continue
            #parse the header names in to (section,name)
            bits = code.split("--")
            if len(bits)==2:
                section, name = bits
                #No parameter should be varied in the old chain
                #but listed as fixed here
                if (section,name) in self.pipeline.fixed_params:
                    logs.error("WARNING: %s varied in old chain now fixed.  I will fix it <--------- Read this warning." % name)
                elif (section,name) in self.pipeline.varied_params:
                    #Record the values of this parameter for later importance
                    #sampling
                    logs.overview(f"Found column in both pipelines {code}:")
                    self.samples.append(col)
                    self.varied_params.append(bits)
                else:
                    logs.overview(f"Found column just in old pipeline: {code}")
                    #This parameter was varied in the old code but is not
                    #here.  So we just save it for output
                    self.original_extras[code] = col
                    self.output.add_column(code, float)
            # anything here must be a sampler-specific 
            else:
                logs.overview(f"Found non-parameter column: {code}")
                self.original_extras[code] = col
                if code=="weight":
                    code="old_weight"
                    logs.overview("Renaming weight -> old_weight")
                self.output.add_column(code, float)

        #Now finally add our actual two sampler outputs, old_like and like
        #P is the original likelihood in the old chain we have samples from
        #P' is the new likelihood.
        self.output.add_column("old_post", float) #This is the old likelihood, log(P)
        self.output.add_column("log_weight", float) #This is the log-weight, the ratio of the likelihoods
        self.output.add_column("post", float) #This is the new likelihood, log(P')


        #Now we need to reorder the list to match the order our pipeline is expecting
        #If a parameter is not listed we use the starting value
        reordered_cols = []
        for p in self.pipeline.varied_params:
            try:
                i = self.varied_params.index(p)
            except ValueError:
                i=-1
            if i>=0:
                col = self.samples[i]
            else:
                col = np.repeat(p.start,self.number_samples)
            reordered_cols.append(col)
        self.samples = np.transpose(reordered_cols)

        self.current_index = 0

    def execute(self):
        self.output.comment("Importance sampling from %s"%self.input_filename)

        #Pick out a chunk of samples to run on
        start = self.current_index
        end = start+self.nstep
        samples_chunk = self.samples[start:end]

        #Run the pipeline on each of the samples
        if self.pool:
            results = self.pool.map(task, samples_chunk)
        else:
            results = list(map(task, samples_chunk))

        #Collect together and output the results
        for i,(sample, (new_like, extra)) in enumerate(zip(samples_chunk, results)):
            #We already (may) have some extra values from the pipeline
            #as derived parameters.  Add to those any parameters used in the
            #old pipeline but not the new one
            extra = list(extra)
            for col in list(self.original_extras.values()):
                extra.append(col[start+i])
            #and then the old and new likelihoods
            old_like = self.original_likelihoods[start+i]
            if self.add_to_likelihood:
                new_like += old_like
            weight = new_like-old_like
            extra = list(extra) + [old_like,weight,new_like]
            #and save results
            self.output.parameters(sample, extra)
        #Update the current index
        self.current_index+=self.nstep

    def is_converged(self):
        return self.current_index>=self.number_samples



"""
This is an implementation of the "Snake" sampler from
Mikkelsen et al
http://arxiv.org/pdf/1211.3126.pdf

"""

import numpy as np
import random
import heapq
import sys

class Snake(object):
	def __init__(self, posterior, origin, spacing, threshold, pool=None):
		self.posterior=posterior
		self.origin=np.array(origin)
		self.spacing=np.array(spacing)
		self.ndim = len(origin)
		self.threshold = threshold
		self.best_fit=tuple([0 for i in range(self.ndim)])
		#Temporarily set the pool to None because we have
		#to evaluate the starting point
		self.pool = None
		self.best_fit_like = self.evaluate([self.best_fit])[1][0]
		self.has_blobs = hasattr(self.best_fit_like, "__len__")
		if self.has_blobs:
			self.best_fit_like, _ = self.best_fit_like
		self.likelihoods = {self.best_fit:self.best_fit_like}
		self.internal = set()
		self.surface = [(-self.best_fit_like,self.best_fit)]
		self.best_like_ever = self.best_fit_like
		self.iterations = 0
		self.pool=pool

	def has_adjacent(self, p):
		return self.adjacent_points(p, True)

	def adjacent_points(self, p, has=False):
		p = np.array(p, dtype=int)
		points = []
		#Loop through each dimension, checking
		#if adjacent points in each direction have been evaluated.
		for i in range(self.ndim):
			dx = np.zeros(self.ndim, dtype=int)
			dx[i] = 1
			# Check point in the +ve direction
			q = tuple(p+dx)
			if q not in self.likelihoods:
				if has: return True
				points.append(q)
			# Check point in the -ve direction
			q = tuple(p-dx)
			if q not in self.likelihoods:
				if has: return True
				points.append(q)
		if has: return False
		return points

	def find_best_fit(self):
		go_on = True
		while go_on:
			best_fit_like, best_fit = heapq.nsmallest(1, self.surface)[0]
			if self.has_adjacent(best_fit):
				go_on=False
			else:
				heapq.heappop(self.surface)
		self.best_fit = best_fit
		self.best_fit_like = -best_fit_like
		if self.best_fit_like>self.best_like_ever:
			self.best_like_ever = self.best_fit_like

	def converged(self):
		return (self.best_like_ever - self.best_fit_like > self.threshold)


	def iterate(self):
		self.iterations += 1
		#take the adjacent points from the best fit
		adjacent = self.adjacent_points(self.best_fit)
		#PARALLEL: take more random choices here
		if self.pool is None:
			n = 1
		else:
			n = min(self.pool.size, len(adjacent))

		P = random.sample(adjacent,n)

		#PARALLEL: do the full evaluation here
		X, outputs = self.evaluate(P)
		if self.has_blobs:
			likelihoods = [L[0] for L in outputs]
			blobs = [L[1] for L in outputs]
		else:
			likelihoods = [L for L in outputs]
			blobs = [None for i in range(len(outputs))]

		for (p,L) in zip(P, likelihoods):
			self.likelihoods[p] = L

		#We cannot combine this with the loop
		#above since we then we will not have filled
		#in the likelihoods for all the points nearby
		for (p,L) in zip(P, likelihoods):
			if self.has_adjacent(p):
				heapq.heappush(self.surface, (-L,p))
		
		if not self.has_adjacent(self.best_fit):
			#This point should now move off the surface
			#and into the interior.
			#This should remove the best fit
			self.surface.remove((-self.best_fit_like, self.best_fit))
			heapq.heapify(self.surface)
		#Definitely overkill to do this every time,
		#but better to be safe and as a nice by-product
		#it gives us a chance to do a convergence check
		#cleanly
		self.find_best_fit()
		return X, likelihoods, blobs

	def evaluate(self, indices):
		points = [self.origin + np.array(index)*self.spacing for index in indices]
		if self.pool is None:
			output = list(map(self.posterior, points))
		else:
			output = self.pool.map(self.posterior, points)
		return points, output


def test_like(x):
	L =  -0.5 * ((x[0]-0.5)**2 + (x[1]-0.8)**2)/0.02**2
	return L

def test():
	snake = Snake(test_like, [0.0, 0.0], [0.01, 0.01], None)
	while not snake.converged():
		snake.iterate()

	for p, L in list(snake.likelihoods.items()):
		print(p[0], p[1], L)
	sys.stderr.write("Ran for %d iterations\n"%snake.iterations)
if __name__ == '__main__':
	test()


from .. import ParallelSampler
from ...runtime import logs
import numpy as np
from . import metropolis
from cosmosis.runtime.analytics import Analytics
import os

#We need a global pipeline
#object for MPI to work properly
pipeline=None


METROPOLIS_INI_SECTION = "metropolis"

def posterior(p):
    return pipeline.run_results(p)


class MetropolisSampler(ParallelSampler):
    parallel_output = True
    sampler_outputs = [("prior", float),("post", float)]
    understands_fast_subspaces = True
    supports_resume = True

    def config(self):
        global pipeline
        pipeline = self.pipeline
        self.samples = self.read_ini("samples", int, default=20000)
        random_start = self.read_ini("random_start", bool, default=False)
        covmat_sample_start = self.read_ini("covmat_sample_start", bool, default=False)
        use_cobaya = self.read_ini("cobaya", bool, default=False)
        self.Rconverge = self.read_ini("Rconverge", float, -1.0)
        self.drag = self.read_ini("drag", int, 0)
        self.oversampling = self.read_ini("oversampling", int, 5)
        tuning_frequency = self.read_ini("tuning_frequency", int, -1)
        tuning_grace = self.read_ini("tuning_grace", int, 5000)
        self.tuning_end = self.read_ini("tuning_end", int, 100000)
        self.save_during_tuning = self.read_ini("save_during_tuning", bool, False)
        self.n = self.read_ini("nsteps", int, default=100)
        self.exponential_probability = self.read_ini("exponential_probability", float, default=0.333)
        self.split = None #work out later
        if self.Rconverge==-1.0:
            self.Rconverge=None
        self.interrupted = False
        self.num_samples = 0
        self.ndim = len(self.pipeline.varied_params)
        self.num_samples_post_tuning = 0
        self.last_accept_count = 0
        #Any other options go here

        # if we are not tunning then there is no tuning phase
        if tuning_frequency == -1:
            self.tuning_end = 0

        if (self.drag > 0) and not self.pipeline.do_fast_slow:
            logs.warning("You asked for dragging, but the pipeline does not have fast/slow enabled"
                  ", so no draggng will be done."
                )

        if (self.pipeline.do_fast_slow) and not (self.pipeline.first_fast_module):
            raise ValueError("To use fast/slow splitting with metropolis please "
                             "manually define first_fast_module in the pipeline "
                             "section.")


        #Covariance matrix
        covmat = self.load_covariance_matrix()

        #start values from prior
        start = self.define_parameters(random_start, covmat_sample_start, covmat)
        logs.overview("MCMC starting point:")
        for param, x in zip(self.pipeline.varied_params, start):
            logs.overview(f"    {param}  {x}")


        if use_cobaya:
            logs.overview("Using the Cobaya proposal")

        logs.overview(f"Will tune every {tuning_frequency} samples, from samples "
              f"{tuning_grace} to {self.tuning_end}.")

        self.sampler = metropolis.MCMC(start, posterior, covmat,
            tuning_frequency=tuning_frequency, # Will be multiplied by the oversampling
            tuning_grace=tuning_grace,         # within the sampler if needed
            tuning_end=self.tuning_end,
            exponential_probability=self.exponential_probability,
            use_cobaya=use_cobaya,
            n_drag = self.drag,
        )
        self.analytics = Analytics(self.pipeline.varied_params, self.pool)
        self.fast_slow_done = False

    def worker(self):
        while not self.is_converged():
            self.execute()
            if self.output:
                self.output.flush()


    def resume(self):
        resume_info = self.read_resume_info()
        if resume_info is None:
            return

        self.sampler, self.num_samples, self.num_samples_post_tuning = resume_info

        # Fast slow is already configured on the sampler.
        self.fast_slow_done = True

        # If we started main sampling (as opposed to tuning phase)
        # then we will have some existing chain, but this is not always the case
        try:
            data = np.genfromtxt(self.output._filename, invalid_raise=False)[:, :self.ndim]
            self.analytics.add_traces(data)
        except IndexError:
            data = None

        
        if self.num_samples >= self.samples:
            logs.important("You told me to resume the chain - it has already completed (with {} samples), so sampling will end.".format(len(data)))
            logs.important("Increase the 'samples' parameter to keep going.")
        elif self.is_converged():
            logs.overview("The resumed chain was already converged.  You can change the converged testing parameters to extend it.")
        elif data is None:
            logs.overview("Continuing metropolis from existing chain - you were in the tuning phase, which will continue")
        else:
            logs.overview("Continuing metropolis from existing chain - have {} samples already".format(len(data)))






    def execute(self):
        #Run the MCMC  sampler.
        if self.pipeline.do_fast_slow and not self.fast_slow_done:
            self.fast_slow_done = True
            self.sampler.set_fast_slow(
                self.pipeline.fast_param_indices,
                self.pipeline.slow_param_indices, 
                self.oversampling
            )

        try:
            samples = self.sampler.sample(self.n)
        except KeyboardInterrupt:
            self.interrupted=True
            return
        self.num_samples += self.n
        self.num_samples_post_tuning = self.num_samples - self.tuning_end


        overall_rate = (self.sampler.accepted * 1.0) / self.sampler.iterations
        recent_accepted = self.sampler.accepted - self.last_accept_count
        recent_rate = recent_accepted / self.n
        logs.overview("Overall accepted {} / {} samples ({:.1%})" .format(
            self.sampler.accepted, self.sampler.iterations, overall_rate))
        logs.overview("Last {0} accepted {1} / {0} samples ({2:.1%})\n" .format(
            self.n, recent_accepted, recent_rate))
        self.last_accept_count = self.sampler.accepted

        # Regardless of save settings we never use tuning samples
        # for analytics
        if self.num_samples_post_tuning > 0:
            traces = np.array([r.vector for r in samples[-self.num_samples_post_tuning:]])
            self.analytics.add_traces(traces)


        if (self.num_samples_post_tuning > 0) or self.save_during_tuning:
            for i, result in enumerate(samples):
                self.output.parameters(result.vector, result.extra, result.prior, result.post)

        if self.num_samples_post_tuning <= 0:
            logs.overview("Tuning ends at {} samples\n".format(self.tuning_end))

        self.write_resume_info([self.sampler, self.num_samples, self.num_samples_post_tuning])

    def is_converged(self):
         # user has pressed Ctrl-C
        if self.interrupted:
            return True
        if self.num_samples >= self.samples:
            rank = self.pool.rank if self.pool is not None else 0
            logs.overview(f"Rank {rank}: Full number of samples generated; sampling complete")
            return True
        elif (self.num_samples > 0 and
              self.pool is not None and
              self.Rconverge is not None and
              self.num_samples_post_tuning > 0):
            R = self.analytics.gelman_rubin()
            R1 = abs(R - 1)
            return np.all(R1 <= self.Rconverge)
        else:
            return False



    def load_covariance_matrix(self):
        covmat_filename = self.read_ini("covmat", str, "").strip()
        if covmat_filename == "" and self.distribution_hints.has_cov():
                covmat =  self.distribution_hints.get_cov()
                logs.overview("Using covariance from previous sampler")
        elif covmat_filename == "":
            logs.overview("Using default covariance 1% of param widths")
            covmat = np.array([p.width()/100.0 for p in self.pipeline.varied_params])
        elif not os.path.exists(covmat_filename):
            raise ValueError(
            "Covariance matrix %s not found" % covmat_filename)
        else:
            logs.overview("Loading covariance from {}".format(covmat_filename))
            covmat = np.loadtxt(covmat_filename)

        if covmat.ndim == 0:
            covmat = covmat.reshape((1, 1))
        elif covmat.ndim == 1:
            covmat = np.diag(covmat ** 2)

        nparams = len(self.pipeline.varied_params)
        if covmat.shape != (nparams, nparams):
            raise ValueError("The covariance matrix was shape (%d x %d), "
                    "but there are %d varied parameters." %
                    (covmat.shape[0], covmat.shape[1], nparams))
        return covmat



    def define_parameters(self, random_start, covmat_sample_start, covmat):
        if covmat_sample_start:
            p = self.pipeline.start_vector()
            chol = np.linalg.cholesky(covmat)

            # Try 100 times to get points within the prior
            for i in range(200):
                r = np.random.normal(size=covmat.shape[0])
                start = p + chol @ r
                prior = self.pipeline.prior(start, total_only=True)
                if np.isfinite(prior):
                    break
            else:
                raise ValueError("You set covmat_sample_start=T so I tried "
                                 "100 times to get a sample inside the prior, "
                                 "but it was always outside."
                    )
            return start
        elif random_start:
            return self.pipeline.randomized_start()
        else:
            return self.pipeline.start_vector()



import numpy as np
from .. import ParallelSampler
from ...runtime import logs

def task(p):
    i,p = p
    r = sampler.pipeline.run_results(p)
    #If requested, save the data to file
    if sampler.save_name:
        if r.block is None:
            logs.error("Failed to run parameters: {} so not saving".format(p))
        else:
            filename = "{}_{}".format(sampler.save_name, i)
            r.block.save_to_file(filename, clobber=True)

    return (r.prior, r.post, r.extra)




class AprioriSampler(ParallelSampler):
    parallel_output = False
    supports_resume = True
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global sampler
        sampler = self

        self.converged = False
        self.save_name = self.read_ini("save", str, "")
        self.nsample = self.read_ini("nsample", int, 1)
        self.n = 0

    def resume(self):
        if self.output.resumed:
            data = np.genfromtxt(self.output._filename, invalid_raise=False)
            self.n = len(data)
            if self.n >= self.nsample:
                logs.error(f"You told me to resume the apriori sampler - it has already completed (with {self.n} samples), so sampling will end.")
                logs.error("Increase the 'nsample' parameter to keep going.")
            else:
                logs.overview(f"Continuing apriori sampling - have {self.n} samples already")


    def execute(self):
        n = self.n
        nparam = len(self.pipeline.varied_params)

        if self.pool:
            chunk_size = self.pool.size

        def sample_from_prior():
            x = np.random.uniform(0.0,1.0,size=nparam)
            p = self.pipeline.denormalize_vector_from_prior(x)
            return p

        while n < self.nsample:
            if self.pool:
                jobs = [(n+i,sample_from_prior()) for i in range(chunk_size)]
                results = self.pool.map(task, jobs)
                n += chunk_size
            else:
                jobs = [(n,sample_from_prior())]
                results = list(map(task, jobs))
                n += 1

            #Save the results of the sampling
            #We now need to abuse the output code a little.
            for sample, result  in zip(jobs, results):
                i, sample=sample
                #Optionally save all the results calculated by each
                #pipeline run to files
                (prior, post, extra) = result
                #always save the usual text output
                self.output.parameters(sample, extra, prior, post)

        #We only ever run this once, though that could 
        #change if we decide to split up the runs
        self.converged = True

    def is_converged(self):
        return self.converged
